(in-package cl-user)
(eval-when (:compile-toplevel :load-toplevel :execute)
  (asdf:operate 'asdf:load-op 'sb-rt))

(declaim (optimize (speed 3)(safety 0) (debug 0)))

(defun time-sample-form (form &optional &key (samples 10))
  (let (start end times)
    (dotimes (i samples)
      (setq start (get-internal-real-time))
      (funcall form)
      (setq end (get-internal-real-time))
      (push (- end start) times))
    
    (flet ((calc-avg (list) (float (/ (apply #'+ list) (length list))))
	   (sq (x) (* x x)))
      
      (let* ((avg (calc-avg times))
	   (sq-times (mapcar #'sq times))
	   (stddev (sqrt (- (calc-avg sq-times) (sq (calc-avg times))))))
;;      (format t "; times ~S, sqtimes ~S~%" times sq-times)
      (format t "; ~D samples, avg ~5F sec, stddev ~5F sec~%" 
	      samples 
	      (/ avg internal-time-units-per-second)
	      (/ stddev internal-time-units-per-second))))))

(defmacro mref (m i j)
    "column i row j"
    `(aref ,m (+ (* ,j 4) ,i)))

(defun make-empty-matrix(length)
  "Makes a 4x4 Opengl Compatible Matrix"
  (make-array length :element-type 'single-float))


(defun mat-mat-mult(a   b  c)
  (declare (type (simple-array single-float (16)) a b c) )
  (loop for row fixnum from 0 to 3 do
          (loop for col fixnum from 0 to 3 do
                (setf (mref c col row)
                      (loop for i fixnum  from 0 to 3
                            sum (*  (mref a col i)
                                    (mref b i row)))))))
(defun random-matrix (length)
  (let ((m (make-array length :element-type 'single-float)))
    (dotimes (i length)
      (setf (aref m i)(random 1000.0)))
    m))

(defun mat-vec-mult (m v r)
  (loop for i from 0 to 3 do
        (setf (aref r i)
              (loop for j from 0 to 3
                    sum (* (mref m i j)
                           (aref v j))))))

(defun nearly-equal (a b)
  (declare (type float a b))
  (< (abs (- a b))  (* 0.0001 a)))

(defun vector-equal (a b)
    (loop for n across a 
          for s across b
          always (nearly-equal n s)))

(sb-rt:deftest nearly-equal
    (and (nearly-equal 1.0 1.00001)
         (nearly-equal 200000.0 200001.0)
         (not (nearly-equal 200000.0 201000.0)))
  t)
(sb-rt:deftest matrix-vector-mult
  (loop for i from 1 to 1000
        always (let ((m (random-matrix 16))
                     (v (random-matrix 4))
                     (r-1 (make-empty-matrix 4))
                     (r-2 (make-empty-matrix 4)))
                 (sb-sys:%primitive sb-vm::matrix-vector-mult m v r-1)
                 (mat-vec-mult m v r-2)
                 (vector-equal r-1 r-2)))
  t)


(sb-rt:deftest matrix-matrix-mult
    (loop for i from 1 to 1000
          always (let* ((m (random-matrix 16))
                        (v (random-matrix 16))
                        (r-1 (make-empty-matrix 16))
                        (r-2 (make-empty-matrix 16 )))
                   (mat-mat-mult m v r-2)
                   (sb-sys:%primitive sb-vm::matrix-matrix-mult m v r-1)
                   (vector-equal r-1 r-2)))
  t)

(defun matrix-vector-timing-test (loops-per-sample samples)
  (let* ((m (random-matrix 16))
         (v (random-matrix 4))
         (r-1 (make-empty-matrix 4)))
    (format t "~D loops per sample ~%" loops-per-sample)
    (format t "sse")
    (time-sample-form #'(lambda ()
                          (dotimes (i loops-per-sample)(sb-sys:%primitive sb-vm::matrix-vector-mult m v r-1)))
                      :samples samples)
    (format t "common-lisp")
    (time-sample-form #'(lambda ()
                          (dotimes (i loops-per-sample)(mat-vec-mult m v r-1)))
                      :samples samples)))

(defun matrix-matrix-timing-test (loops-per-sample samples)
  (let* ((m (random-matrix 16))
         (v (random-matrix 16))
         (r-1 (make-empty-matrix 16)))
    (format t "~D loops per sample ~%" loops-per-sample)
    (format t "sse")
    (time-sample-form #'(lambda ()
                          (dotimes (i loops-per-sample)(sb-sys:%primitive sb-vm::matrix-matrix-mult m v r-1)))
                      :samples samples)
    (format t "common-lisp")
    (time-sample-form #'(lambda ()
                          (dotimes (i loops-per-sample)(mat-mat-mult m v r-1)))
                      :samples samples)))
                   
;; CL-USER> (matrix-matrix-timing-test 100000 1000)
;; 100000 loops per sample 
;; sse; 1000 samples, avg 0.013 sec, stddev .0005 sec
;; common-lisp; 1000 samples, avg 1.294 sec, stddev .0215 sec


